// The MIT License
//
// Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Copyright (C) 2020 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import Foundation
import DependencyInjection
import Logger
import UIKit
import Combine
import ConfigurationServices
import LocalAuthentication
import HTTPNetworking
import SharedServices
import UserSettingsServices
import MaternityServices
import CMSServices

enum BuildConfiguration: String {
    case debug = "Debug"
    case projectEnv = "ProjectEnv"
    case internalTest = "InternalTest"
    case test1 = "Test1"
    case test2 = "Test2"
    case release = "Release"
    case unknown = "Unknown"

    /// This is not a BuildConfiguration we can build as, but a mode we can force the app into, when needed
    /// The App Store reviewer can long press an element on the first screen of the app, and enter this mode to avoid NemID login.
    case appStoreReview = "AppStoreReview"
    
    var isFirebaseDistribution: Bool {
        switch self {
        case .debug, .release, .unknown:
            return false
        case .projectEnv, .internalTest, .test1, .test2, .appStoreReview:
            return true
        }
    }

    /// E.g. ErrorIDs from the App Facade API
    var canShowDebugValues: Bool {
        switch self {
        case .debug, .projectEnv, .internalTest, .test1, .test2:
            return true
        case .release:
            return true // This is currently enabled for the pilot release. We will determine whether this is OK for the final release version too.
        case .appStoreReview, .unknown:
            return false
        }
    }

    /// E.g. stack trace info from the backend
    var canShowDebugDescriptions: Bool {
        switch self {
        case .debug, .projectEnv, .internalTest, .test1:
            return true
        case .test2, .release, .appStoreReview, .unknown:
            return false
        }
    }
}

struct Configuration {
    let useMock: Bool
    let loggerConfig: OSLoggerConfig
}

final class App {
    private (set) static var buildConfiguration = BuildConfiguration(rawValue: Bundle.main.buildConfiguration) ?? .unknown
    private (set) static var configuration = Configuration.build(for: buildConfiguration)
    static let isTesting: Bool = isUnitTesting || isUITesting
    static let isUnitTesting: Bool = ProcessInfo.processInfo.environment["IS_TESTING"] == "1"
    static let isUITesting: Bool = ProcessInfo.processInfo.environment["IS_UI_TESTING"] == "1"
    
    static let preferredLandscapeOrientation: UIInterfaceOrientation = .landscapeRight
    
    private static let multiLogger = MultiLogger(loggers: [])
    
    /// Used to present OpenID Connect login view
    static var topViewController: UIViewController? {
        UIApplication.topViewController()
    }
    
    static func buildDependencies() {
        setupLoggers()
        setupMockFailRisk()
        
        addUtilityDependencies()
        addUserDependencies()
        addBackendForFrontendConfigurations()
        addAppleDependencies()
        addCMSDependencies()
        addMaternityDependencies()
        addConfigurationDependencies()
        addRepositories()
        addLocalStorage()
    }

    static func enterAppStoreReviewModeAndRebuildAllDependencies() {
        multiLogger.debug("Entering App Store Review build configuration mode!")
        Self.buildConfiguration = .appStoreReview
        configuration = Configuration.build(for: Self.buildConfiguration)
        Resolver.shared.allowOverride = true
        buildDependencies()
    }
    
    private static func addAppleDependencies() {
        Resolver.shared.register(LocalAuthenticationContext.self, { LAContext() })
        Resolver.shared.registerAsSingleton(NetworkSession.self, URLSession.shared)
    }
    
    private static func addCMSDependencies() {
        Resolver.shared.registerAsSingleton(CMSApiConfiguration.self, CMSConfiguration())
        
        if configuration.useMock {
            Resolver.shared.registerAsSingleton(DictionaryService.self, DictionaryServiceMock())
            Resolver.shared.registerAsSingleton(ChecklistContentService.self, ChecklistContentServiceMock())
            Resolver.shared.registerAsSingleton(PagesService.self, PagesServiceMock())
            Resolver.shared.registerAsSingleton(FAQService.self, FAQServiceMock())
            Resolver.shared.registerAsSingleton(ArticlesService.self, ArticlesServiceMock())
        } else {
            Resolver.shared.registerAsSingleton(DictionaryService.self, DictionaryServiceImpl())
            Resolver.shared.registerAsSingleton(ChecklistContentService.self, ChecklistContentServiceImpl())
            Resolver.shared.registerAsSingleton(PagesService.self, PagesServiceImpl())
            Resolver.shared.registerAsSingleton(FAQService.self, FAQServiceImpl())
            Resolver.shared.registerAsSingleton(ArticlesService.self, ArticlesServiceImpl())
        }
    }
    
    private static func addBackendForFrontendConfigurations() {
        let apiConfiguration = ApiConfiguration()
        Resolver.shared.registerAsSingleton(ConfigurationApiConfiguration.self, apiConfiguration)
    }
    
    private static func addMaternityDependencies() {
        if configuration.useMock {
            Resolver.shared.registerAsSingleton(CarePlanService.self, CarePlanServiceMock())
        } else {
            Resolver.shared.registerAsSingleton(CarePlanService.self, CarePlanServiceImpl())

        }
    }
    
    private static func addConfigurationDependencies() {
        if configuration.useMock {
            Resolver.shared.registerAsSingleton(ConfigurationService.self, ConfigurationServiceMock())
        } else {
            Resolver.shared.registerAsSingleton(ConfigurationService.self, ConfigurationServiceImpl())
        }
    }
    
    private static func addUserDependencies() {
        if configuration.useMock {
            Resolver.shared.registerAsSingleton(UserSettingsService.self, UserSettingsServiceMock(isPersistant: !isTesting))
        } else {
            Resolver.shared.registerAsSingleton(UserSettingsService.self, UserSettingsServiceImpl())
        }
        Resolver.shared.registerAsSingleton(UserContextProtocol.self, UserContext())
    }
    
    private static func addUtilityDependencies() {
        Resolver.shared.registerAsSingleton(Logger.self, multiLogger)
        
        if configuration.useMock {
            Resolver.shared.registerAsSingleton(SystemInformationService.self, SystemInformationServiceMock())
        } else {
            Resolver.shared.registerAsSingleton(SystemInformationService.self, SystemInformationServiceImpl())
        }
        Resolver.shared.registerAsSingleton(ChecklistRepository.self, ChecklistRepositoryImpl())
    }
    
    private static func addRepositories() {
        Resolver.shared.registerAsSingleton(MaternityRepository.self, MaternityRepositoryImpl.shared)
    }
    
    private static func addLocalStorage() {
        Resolver.shared.registerAsSingleton(MaternityStorageService.self, MaternityStorageServiceImpl.shared)
        Resolver.shared.registerAsSingleton(ChecklistStorageService.self, ChecklistStorageServiceImpl.shared)
    }
    
    private static func setupLoggers() {
        var loggers: [LoggerProtocol] = [OSLogger(config: configuration.loggerConfig)]
        if buildConfiguration == .debug {
            HTTPNetworkingConfig.isDebugPrintingEnabled = false // Can be enabled when debugging API requests
            loggers.append(ErrorFlashLogger())
        }
        multiLogger.loggers = loggers
    }
    
    static func setupMockFailRisk() {
        if buildConfiguration == .debug && !isTesting {
            MockFlags.failRisk = .small
        } else {
            MockFlags.failRisk = .disabled
        }
    }
}

private extension Configuration {
    //swiftlint:disable:next function_body_length
    static func build(for buildConfiguration: BuildConfiguration) -> Configuration {
        
        switch buildConfiguration {
        case .debug:
            return Configuration(
                useMock: App.isUITesting,
                loggerConfig: OSLoggerConfig(
                    defaultCategory: buildConfiguration.rawValue,
                    minimumLogLevel: .debug,
                    isDevelopmentInfoEnabled: false,
                    isEmojisEnabled: true
                )
            )
        case .projectEnv,
             .internalTest,
             .test1,
             .test2:
            return Configuration(
                useMock: false,
                loggerConfig: OSLoggerConfig(
                    defaultCategory: buildConfiguration.rawValue,
                    minimumLogLevel: .info,
                    isDevelopmentInfoEnabled: true,
                    isEmojisEnabled: true
                )
            )
        case .release:
            return Configuration(
                useMock: false,
                loggerConfig: OSLoggerConfig(
                    defaultCategory: buildConfiguration.rawValue,
                    minimumLogLevel: .error,
                    isDevelopmentInfoEnabled: true,
                    isEmojisEnabled: true
                )
            )
        case .appStoreReview:
            return Configuration(
                useMock: true,
                loggerConfig: OSLoggerConfig(
                    defaultCategory: buildConfiguration.rawValue,
                    minimumLogLevel: .warning,
                    isDevelopmentInfoEnabled: false,
                    isEmojisEnabled: true
                )
            )
        case .unknown:
            fatalError("App was started with unknown build configuration 🤯")
        }
    }
}
