// The MIT License
//
// Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Copyright (C) 2020 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import UIKit
import SwiftUI
import CMSServices
import UserSettingsServices
import LocalAuthentication

final class SceneDelegate: UIResponder, UIWindowSceneDelegate {

    var window: UIWindow?

    private var cancelBag = CancelBag()

    @DI private var logger: Logger
    @DI private var userContext: UserContextProtocol
    @DI private var dictionaryService: DictionaryService
    @DI private var userSettings: UserSettingsService
    @DI private var maternityStore: MaternityRepository

    func scene(_ scene: UIScene, willConnectTo session: UISceneSession, options connectionOptions: UIScene.ConnectionOptions) {
        guard let windowScene = scene as? UIWindowScene, !App.isUnitTesting else { return }

        // Swizzle `UIColor.systemBackground` to change the default background color system-wide.
        // We've have tried to solve this in several different ways, but this is the only way, that
        // doesn't ruins other UI behaviours while setting the background (`UIHostingController` is the big scapegoat in this case 🐐)
        UIColor.swizzleSystemBackground()

        if App.isUITesting {
            clearAllUserDataForUITests()
        }

        let window = UIWindow(windowScene: windowScene)
        self.window = window

        listenForNotifications()

        if !userContext.hasHandledOnboardingFlow {
            // User hasn't finished onboarding
            let onboarding = UIHostingController(rootView: OnboardingView())
            changeRootViewController(onboarding, animated: false)
            
        } else if userContext.hasEnabledAuthentication {
            let storyboard = UIStoryboard(name: R.storyboard.launchScreen.name, bundle: nil)
            let launchScreen = storyboard.instantiateViewController(withIdentifier: "launchScreen")
            changeRootViewController(launchScreen, animated: false)
            authenticate()
        
        } else if userContext.hasHandledOnboardingFlow {
            maternityStore.requestUpdate()
                .sink(receiveCompletion: { _ in
                    self.handleNotification(.showTabBarController, animated: false)},
                      receiveValue: { _ in })
                .store(in: &cancelBag)
        }
        window.makeKeyAndVisible()
        
        if App.isUITesting {
            // Disable animations while testing.
            // This only works on UIKit animations, there is currently no way to disable all SwiftUI animations 😬
            UIView.setAnimationsEnabled(false)
        }
    }

    func sceneDidDisconnect(_ scene: UIScene) {
        // Called as the scene is being released by the system.
        // This occurs shortly after the scene enters the background, or when its session is discarded.
        // Release any resources associated with this scene that can be re-created the next time the scene connects.
        // The scene may re-connect later, as its session was not necessarily discarded (see `application:didDiscardSceneSessions` instead).
    }
    
    func sceneDidBecomeActive(_ scene: UIScene) {
        SecurityCoverView.shared.detachFromWindow()
        
        if App.isUnitTesting {
            // Skip the code block when unit testing
            return
        }

        if !userContext.hasEnabledAuthentication {
            // If authentication is not enabled, just refresh data
            refreshData()
        } else if let lastOpened = UserDefaults.standard.object(forKey: "LastOpened") as? Date,
                  lastOpened.date(byAdding: .second, value: 20)! < Date() {
            // If authentication is enabled and the condition is met, show the launch screen and authenticate
            let storyboard = UIStoryboard(name: R.storyboard.launchScreen.name, bundle: nil)
            let launchScreen = storyboard.instantiateViewController(withIdentifier: "launchScreen")
            changeRootViewController(launchScreen, animated: false)
            authenticate()
        } else {
            // If authentication is enabled but the condition is not met, refresh data fetches
            refreshData()
        }
    }

    func sceneWillResignActive(_ scene: UIScene) {
        guard !App.isTesting else {
            return
        }
        UserDefaults.standard.set(Date(), forKey: "LastOpened")
    }

    func sceneWillEnterForeground(_ scene: UIScene) {
        // Called as the scene transitions from the background to the foreground.
        // Use this method to undo the changes made on entering the background.
    }

    func sceneDidEnterBackground(_ scene: UIScene) {
        guard !App.isTesting else {
            return
        }
        SecurityCoverView.shared.attachToWindow()
    }

    private func refreshData() {
        maternityStore.requestUpdate()
            .sink(receiveCompletion: { _ in }, receiveValue: { _ in })
            .store(in: &cancelBag)
    }
    
    private func listenForNotifications() {
        NotificationCenter.default
            .listen(for: Notification.Name.SceneDelegate.changeRootViewController)
            .sink(receiveValue: { (message: NotificationMessage) in
                self.handleNotification(message, animated: true)
            })
            .store(in: &cancelBag)
    }

    private func changeRootViewController(_ rootViewController: UIViewController, animated: Bool) {
        guard let window = window else {
            fatalError("Failed to change rootViewController, since the windows was nil")
        }
        rootViewController.view.backgroundColor = ColorManager.bgNeutralPrimary.currentUIColor()
        window.rootViewController = rootViewController
        window.makeKeyAndVisible()

        if animated {
            UIView.transition(
                with: window,
                duration: 0.2,
                options: .transitionCrossDissolve,
                animations: {},
                completion: nil
            )

        }
    }

    private func clearAllUserDataForUITests() {
        userSettings.remove(key: .hasHandledOnboarding)
    }
    
    func authenticate() {
        let context = LAContext()
        
        if context.canEvaluatePolicy(.deviceOwnerAuthentication, error: nil) {
            context.evaluatePolicy(.deviceOwnerAuthentication, localizedReason: Localizable.onboarding_login_touch()) { success, _ in
                if success {
                    self.userContext.setHasHandledOnboardingFlow(true)
                    self.maternityStore.requestUpdate()
                        .sink(receiveCompletion: { _ in
                            self.handleNotification(.showTabBarController, animated: false)},
                              receiveValue: { _ in })
                        .store(in: &self.cancelBag)
                }
            }
        }
    }
}

extension SceneDelegate {
    enum NotificationMessage : NotificationCenterValue {
        case showTabBarController
        case resetToOnboarding
    }

    private func handleNotification(_ notification: NotificationMessage, animated: Bool) {
        switch notification {
        case .showTabBarController:
            changeRootViewController(TabBarController(), animated: animated)
        case .resetToOnboarding:
            let onboarding = UIHostingController(rootView: OnboardingView())
            changeRootViewController(onboarding, animated: animated)
        }
    }
}
