// The MIT License
//
// Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Copyright (C) 2020 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of
// this software and associated documentation files (the "Software"), to deal in
// the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

import UIKit
import Charts

final class MeasurementsLineChartRenderer: LineChartRenderer {
    static let dotSize: CGFloat = 12
    static let dotRadius: CGFloat = dotSize / 2

    var dataSets: [ChartDataSetProtocol] {
        dataProvider?.lineData?.dataSets ?? []
    }

    override func drawExtras(context: CGContext) {
        super.drawExtras(context: context)

        for dataSet in dataSets {

            switch dataSet {
            case is MeasurementLineChartValueDataSet:
                for entryIndex in 0 ..< dataSet.entryCount {
                    if let entry = dataSet.entryForIndex(entryIndex) {
                        drawValueCircle(entry: entry, dataSet: dataSet as? ChartDataSet ?? ChartDataSet(entries: [ChartDataEntry]()), in: context)
                    }
                }
            default:
                break // No extras!
            }
        }
    }

    func drawLinear(context: CGContext, dataSet: LineChartDataSet) {

        // Do not draw line for values.
        if let dataSet = dataSet as? MeasurementLineChartAreaDataSet {
            context.saveGState()
            context.setFillColor(ColorManager.iconDecorativePurpleHeart.currentUIColor().withAlphaComponent(0.08).cgColor)
            if let transformer = dataProvider?.getTransformer(forAxis: dataSet.axisDependency) {
                let valuesForArea = [
                    dataSet.upperEntries,
                    dataSet.lowerEntries.reversed()
                ].flatMap({ $0 })

                let bezier = UIBezierPath()

                for areaValue in valuesForArea {
                    let point = CGPoint(x: areaValue.x, y: areaValue.y).applying(transformer.valueToPixelMatrix)
                    if areaValue === valuesForArea.first {
                        bezier.move(to: point)
                    }
                    bezier.addLine(to: point)
                }
                bezier.close()
                bezier.fill()
            }
            context.restoreGState()
        }
        // Draws middle line for area and draws value line for values.
        super.drawLinear(context: context, dataSet: dataSet)
    }

    private func drawValueCircle(entry: ChartDataEntry, dataSet: ChartDataSet, in context: CGContext) {

        guard let transformer = dataProvider?.getTransformer(forAxis: dataSet.axisDependency) else {
            return
        }

        let pixelPoint = CGPoint(x: entry.x, y: entry.y).applying(transformer.valueToPixelMatrix)
        var rect = CGRect(origin: pixelPoint, size: CGSize(width: Self.dotSize, height: Self.dotSize))

        rect.origin.x = pixelPoint.x - Self.dotRadius
        rect.origin.y = pixelPoint.y - Self.dotRadius
        rect.size.width = Self.dotSize
        rect.size.height = Self.dotSize

        context.saveGState()
        context.setStrokeColor(UIColor.white.cgColor)
        context.setLineWidth(5)
        context.setFillColor(ColorManager.iconDecorativePurpleHeart.currentUIColor().cgColor)
        context.setShadow(offset: .zero, blur: 4, color: ColorManager.shadow.currentUIColor().withAlphaComponent(0.5).cgColor)

        context.strokeEllipse(in: rect)
        context.fillEllipse(in: rect)

        context.restoreGState()
    }
}
